@@ -1,6 +1,6 @@
 /*
 	gentree.cc	Generate Directory Tree For Display
-	Copyright (c) 2000 Kriang Lerdsuwanakij
+	Copyright (c) 2000, 2002 Kriang Lerdsuwanakij
 	email:		lerdsuwa@users.sourceforge.net
 
 	This program is free software; you can redistribute it and/or modify
@@ -101,11 +101,31 @@ void	GenerateHTMLStrings()
 	Generate HTML-like codes from the directory tree
 *************************************************************************/
 
-vector<size_t>	barStack;	// Positions of vertical bars
 size_t	htmlRow, htmlCol;
-size_t	curCol;			// curRow not needed
 
-void	NewLineUpdateRowCol()
+class HyperTreeDraw : public HyperDraw {
+	private:
+		vector<size_t>	barStack;	// Positions of vertical bars
+		size_t	curCol;			// curRow not needed
+
+		void	NewLineUpdateRowCol();
+
+		void	GenerateSymLinkHTML(HyperDocument &html, WINDOW *pad,
+					    sptr<DirectoryEntry> &d,
+					    const string &new_dir);
+
+		void	GenerateDirHTML(HyperDocument &html, WINDOW *pad,
+					sptr<DirectoryEntry> &d,
+					const string &new_dir);
+
+		void	RecursiveGenerateHTML(HyperDocument &html, WINDOW *pad,
+					      sptr<DirectoryEntry> &d, 
+					      size_t pos, const string &current_dir);
+	public:
+		void	Draw(HyperDocument &html, WINDOW *pad);
+};
+
+void	HyperTreeDraw::NewLineUpdateRowCol()
 {
 	if (curCol > htmlCol)
 		htmlCol = curCol;
@@ -113,8 +133,9 @@ void	NewLineUpdateRowCol()
 	curCol = 0;
 }
 
-void	GenerateSymLinkHTML(Buffer &buffer, sptr<DirectoryEntry> &d,
-			    const string &new_dir)
+void	HyperTreeDraw::GenerateSymLinkHTML(HyperDocument &html, WINDOW *pad,
+					   sptr<DirectoryEntry> &d,
+					   const string &new_dir)
 {
 	bool	tagBegin = false;
 
@@ -122,95 +143,101 @@ void	GenerateSymLinkHTML(Buffer &buffer, sptr<DirectoryEntry> &d,
 	    && d->isLinkDestUnreadable != 1) {
 
 						// Write tags
-		buffer.Puts("<A NAME=\"");
-		buffer.Puts(QuoteURLChars(new_dir));
-		buffer.Puts("\"></A><A HREF=\"#");
-		buffer.Puts(QuoteURLChars(d->GetLinkDestStr()));
-		buffer.Puts("\"><I>");
+		html.DrawAName(pad, new_dir);
+
+		string s = "#";
+		s += d->GetLinkDestStr();	// Anything after '#' can contain any
+						// char, so quoting URL is not needed.
+		html.DrawAHrefBegin(pad, s);
+
+		html.DrawSetItalic(pad);
 
 		tagBegin = true;
 	}
 
 			    			// Write dir. name
-	buffer.Puts(ExpandEscape(d->GetNameStr()));
+	html.DrawString(pad, d->GetNameStr());
 	curCol += GetStringWidth(d->GetNameStr());
 			
 			    			// Write graphical " -> "
-	buffer.Puts(" &boxh;&rarr; ");
+	html.DrawChar(pad, ' ');
+	html.DrawEntityChar(pad, EID_BOXH);
+	html.DrawEntityChar(pad, EID_RARR);
+	html.DrawChar(pad, ' ');
 	curCol += 4;
 
 			    			// Write linked dir.
-	buffer.Puts(ExpandEscape(d->GetLinkStr()));
+	html.DrawString(pad, d->GetLinkStr());
 	curCol += GetStringWidth(d->GetLinkStr());
 
 	switch(d->isLinkDestSkip) {
 		case 0:
 			if(d->isLinkDestUnreadable) {
-				buffer.Puts(html_unreadable);
+				html.DrawString(pad, html_unreadable);
 				curCol += html_unreadable_width;
 			}
 			break;
 		case 1:
-			buffer.Puts(html_skipped);
+			html.DrawString(pad, html_skipped);
 			curCol += html_skipped_width;
 			break;
 		case 2:
-			buffer.Puts(html_not_in_file);
+			html.DrawString(pad, html_not_in_file);
 			curCol += html_not_in_file_width;
 			break;
 		case 3:
-			buffer.Puts(html_not_found);
+			html.DrawString(pad, html_not_found);
 			curCol += html_not_found_width;
 			break;
 	}
 			
 	if (tagBegin) {
 			    			// End <A HREF=...> tag
-		buffer.Puts("</I></A>");
+		html.DrawClearItalic(pad);
+		html.DrawAEnd(pad);
 	}
 }
 
-void	GenerateDirHTML(Buffer &buffer, sptr<DirectoryEntry> &d,
-			    const string &new_dir)
+void	HyperTreeDraw::GenerateDirHTML(HyperDocument &html, WINDOW *pad,
+				       sptr<DirectoryEntry> &d,
+				       const string &new_dir)
 {
 	if (d->isUnreadable != 1) {		// Exec permission
 						// Write tags
-		buffer.Puts("<A NAME=\"");
-		buffer.Puts(QuoteURLChars(new_dir));
-		buffer.Puts("\"></A><A CALL=\"");
-		buffer.Puts(QuoteURLChars(new_dir));
-		buffer.Puts("\">");
+		html.DrawAName(pad, new_dir);
+		html.DrawACallBegin(pad, new_dir);
 	}
 			    		
 			    			// Write dir. name
-	buffer.Puts(ExpandEscape(d->GetNameStr()));
+	html.DrawString(pad, d->GetNameStr());
 	curCol += GetStringWidth(d->GetNameStr());
 			    		
 	if (d->isUnreadable != 1) {		// Exec permission
 			    			// Dir. can be accessed
 			    				
-			    			// End <A NAME=...> tag
-		buffer.Puts("</A>");
+			    			// End <A CALL=...> tag
+		html.DrawAEnd(pad);
 	}				    		
 }
 
-void	RecursiveGenerateHTML(Buffer &buffer, sptr<DirectoryEntry> &d, 
-			      size_t pos, const string &current_dir)
+void	HyperTreeDraw::RecursiveGenerateHTML(HyperDocument &html, WINDOW *pad, 
+					     sptr<DirectoryEntry> &d, 
+					     size_t pos, const string &current_dir)
 {
 	if (d->isSkip) {
 		if (d->isSkip == 2) {
-			buffer.Puts(html_not_in_file);
+			html.DrawString(pad, html_not_in_file);
 			curCol += html_not_in_file_width;
 			pos += html_not_in_file_width;
 		}
 		else {
-			buffer.Puts(html_skipped);
+			html.DrawString(pad, html_skipped);
 			curCol += html_skipped_width;
 			pos += html_skipped_width;
 		}
 	}
 	if (d->isUnreadable) {
-		buffer.Puts(html_unreadable);
+		html.DrawString(pad, html_unreadable);
 		curCol += html_unreadable_width;
 		pos += html_unreadable_width;
 	}
@@ -228,34 +255,37 @@ void	RecursiveGenerateHTML(Buffer &buffer, sptr<DirectoryEntry> &d,
 			    			// First subdir entry
 			    			// Draw a bar connected to 
 			    			// parent dir
-			buffer.Puts("&boxh;");
+			html.DrawEntityChar(pad, EID_BOXH);
 			curCol++;
 			
 			if (numSubDir == 1) {
 			    			// Only one subdir present
 			    			// Draw a bar
-			    	buffer.Puts("&boxh;");
+				html.DrawEntityChar(pad, EID_BOXH);
 			    	curCol++;
 			}
 			else {
-			    	buffer.Puts("&boxhd;");
+				html.DrawEntityChar(pad, EID_BOXHD);
 			    	curCol++;
 			    	
 				barStack.push_back(pos+1);
 			}
-			buffer.Puts("&boxh;");
+			html.DrawEntityChar(pad, EID_BOXH);
 			curCol++;
 		}
 		else if (curSubDir == numSubDir-1) {
 			    			// Last subdir entry
-			    			
-			buffer.Puts(" &boxur;&boxh;");
+			html.DrawChar(pad, ' ');
+			html.DrawEntityChar(pad, EID_BOXUR);
+			html.DrawEntityChar(pad, EID_BOXH);
 			curCol += 3;
 			
 			barStack.pop_back();
 		}
 		else {				// Middle entries
-			buffer.Puts(" &boxvr;&boxh;");
+			html.DrawChar(pad, ' ');
+			html.DrawEntityChar(pad, EID_BOXVR);
+			html.DrawEntityChar(pad, EID_BOXH);
 			curCol += 3;
 		}
 			    
@@ -266,33 +296,33 @@ void	RecursiveGenerateHTML(Buffer &buffer, sptr<DirectoryEntry> &d,
 
 			    			// For links
 		if ((*iter)->IsSymLink()) {
-			GenerateSymLinkHTML(buffer, (*iter), new_dir);
+			GenerateSymLinkHTML(html, pad, (*iter), new_dir);
 		}
 		else {
-			GenerateDirHTML(buffer, (*iter), new_dir);
+			GenerateDirHTML(html, pad, (*iter), new_dir);
 			
 				// Note: subdir. allowed even inside
 				// unreadable or skipped dir
 			
 			    			// Scan inside this dir
-		    	RecursiveGenerateHTML(buffer, *iter,
+		    	RecursiveGenerateHTML(html, pad, *iter,
 					      pos+3+GetStringWidth((*iter)->GetNameStr()),
 					      new_dir);
 	    	}
 		if (curSubDir != numSubDir-1) {	// Not the last one
 						// New line
-			buffer.Putc('\n');
+			html.DrawChar(pad, '\n');
 			NewLineUpdateRowCol();
 						// Add vertical bars
 			for (size_t i = 0, j = 0; i < pos; i++) {
 				if (j < barStack.size() && barStack[j] == i) {
-					buffer.Puts("&boxv;");
+					html.DrawEntityChar(pad, EID_BOXV);
 					curCol++;
 					
 					j++;
 				}
 				else {
-					buffer.Putc(' ');
+					html.DrawChar(pad, ' ');
 					curCol++;
 				}
 			}
@@ -300,24 +330,29 @@ void	RecursiveGenerateHTML(Buffer &buffer, sptr<DirectoryEntry> &d,
 	}
 }
 
-void	GenerateHTML(Buffer &buffer)
+void	HyperTreeDraw::Draw(HyperDocument &html, WINDOW *pad)
 {
 	htmlRow = 0;
 	htmlCol = 0;
 	curCol = 0;
-	
-	time_t scanTime = GetScanTime();
-	struct tm *scanTime2 = localtime(&scanTime);
 
-	ostrstream bufstr;
-	gtout(bufstr, _("<TITLE>%$ %$   Last scan: %$</TITLE>"))
-	     << progName << version
-	     << my_strftime(_("%b %e  %H:%M"), scanTime2);
-	bufstr << ends;
-	buffer.Puts(bufstr.str());
-	bufstr.freeze(false);
+	if (! pad) {
+		;			// Nothing to do
+	}
+	else {
+		html.DrawBegin(pad);
 
-	buffer.Puts("<PRE>");		// Start preformatted text
+		time_t scanTime = GetScanTime();
+		struct tm *scanTime2 = localtime(&scanTime);
+
+		ostrstream bufstr;
+		gtout(bufstr, _("%$ %$   Last scan: %$"))
+		     << progName << version
+		     << my_strftime(_("%b %e  %H:%M"), scanTime2);
+		bufstr << ends;
+		html.DrawSetTitle(bufstr.str());
+		bufstr.freeze(false);
+	}
 
 	bool	first = true;
 
@@ -325,7 +360,7 @@ void	GenerateHTML(Buffer &buffer)
 	     iter != dirTree.end(); ++iter) {
 	
 		if (!first) {		// New line need ?
-			buffer.Putc('\n');
+			html.DrawChar(pad, '\n');
 			NewLineUpdateRowCol();
 		}
 		else
@@ -334,24 +369,29 @@ void	GenerateHTML(Buffer &buffer)
 		string new_dir = (*iter)->GetNameStr();
 
 		if ((*iter)->IsSymLink()) {		// Symbolic link
-			GenerateSymLinkHTML(buffer, (*iter), new_dir);
+			GenerateSymLinkHTML(html, pad, (*iter), new_dir);
 		}
 		else {
-			GenerateDirHTML(buffer, (*iter), new_dir);
+			GenerateDirHTML(html, pad, (*iter), new_dir);
 
 			curCol = GetStringWidth(new_dir);
 
 				// Note: subdir. allowed even inside
 				// unreadable or skipped dir
 						// Recursion...
-			RecursiveGenerateHTML(buffer, *iter, curCol, new_dir);
+			RecursiveGenerateHTML(html, pad, *iter, curCol, new_dir);
 		}
 	}
 
-	buffer.Puts("</PRE>");		// End preformatted text
-
 	if (curCol != 0)		// Not end with new line
 		NewLineUpdateRowCol();	// One more line needed
+
+	if (! pad) {
+		html.DrawSetSize(htmlRow, htmlCol);
+	}
+	else {
+		;			// Nothing to do
+	}
 }
 
 /*************************************************************************
@@ -425,76 +465,102 @@ size_t	FindMatchDir(list<MatchInfo> &match_list, const string &linktext,
 	return match_list.size();
 }
 
+/*************************************************************************
+	Generate HTML-like codes from the match list
+*************************************************************************/
 
-void	GenerateDirListHTML(list<MatchInfo> &match_list, Buffer &buffer)
+class HyperListDraw : public HyperDraw {
+		size_t		numDigit, max_length;
+		list<MatchInfo> &match_list;
+	public:
+		HyperListDraw(list<MatchInfo> &match_list_);
+		void Draw(HyperDocument &html, WINDOW *pad);
+};
+
+HyperListDraw::HyperListDraw(list<MatchInfo> &match_list_) :
+				match_list(match_list_)
 {
-	htmlRow = 0;
-	htmlCol = 0;
-	curCol = 0;
-	
 	size_t	tempNumMatch = match_list.size();
-	size_t	numDigit = 0;
+	numDigit = 0;
 	while (tempNumMatch) {
 		tempNumMatch /= 10;
 		numDigit++;
 	}
-	
-	time_t scanTime = GetScanTime();
-	struct tm *scanTime2 = localtime(&scanTime);
 
-	ostrstream bufstr;
-	gtout(bufstr, _("<TITLE>%$ %$   Found: %$   Last scan: %$</TITLE>"))
-	     << progName << version << match_list.size()
-	     << my_strftime(_("%b %e  %H:%M"), scanTime2);
-	bufstr << ends;
-	buffer.Puts(bufstr.str());
-	bufstr.freeze(false);
-
-	buffer.Puts("<PRE>");		// Start preformatted text
-
-	size_t	index = 1;
-
-	size_t	max_length = 0;
+	max_length = 0;
 	for (list<MatchInfo>::iterator iter = match_list.begin();
 	     iter != match_list.end(); iter++) {
 		size_t size = GetStringWidth(iter->name);
 		if (size > max_length)
 			max_length = size;
 	}
+}
 
-	for (list<MatchInfo>::iterator iter = match_list.begin();
-	     iter != match_list.end(); iter++) {
+void	HyperListDraw::Draw(HyperDocument &html, WINDOW *pad)
+{
+
+	if (!pad) {
+				// Determine document size
+//		ostrstream bufstr;
+//		gtout(bufstr, "echo %$ %$ > xx") << match_list.size() << 3+numDigit+max_length;
+//		bufstr << ends;
+//		system(bufstr.str());
+//		bufstr.freeze(false);
+
+		html.DrawSetSize(match_list.size(), 3+numDigit+max_length);
+	}
+	else {
+				// Draw document
+		html.DrawBegin(pad);
+
+		string	s = "";
+		for (size_t i = 0; i < numDigit + 2; ++i)
+			s += ' ';
+
+		time_t scanTime = GetScanTime();
+		struct tm *scanTime2 = localtime(&scanTime);
+
+		ostrstream bufstr;
+		gtout(bufstr, _("%$ %$   Found: %$   Last scan: %$"))
+		     << progName << version << match_list.size()
+		     << my_strftime(_("%b %e  %H:%M"), scanTime2);
+		bufstr << ends;
+		html.DrawSetTitle(bufstr.str());
+		bufstr.freeze(false);
+
+		size_t	index = 1;
+		for (list<MatchInfo>::iterator iter = match_list.begin();
+		     iter != match_list.end(); iter++) {
 	
-		if (index != 1) {		// New line need ?
-			buffer.Putc('\n');
-			NewLineUpdateRowCol();
-		}
+			if (index != 1) {		// New line need ?
+				html.DrawChar(pad, '\n');
+			}
 
 					// Write tags
-		buffer.Puts("<A NAME=\"");
-		buffer.Puts(QuoteURLChars(iter->name));
-		buffer.Puts("\"></A><A CALL=\"");
-		buffer.Puts(QuoteURLChars(iter->name));
-		buffer.Puts("\">");
+			html.DrawAName(pad, iter->name);
+			html.DrawACallBegin(pad, iter->name);
 
-		buffer.Printf(" %*d ", numDigit, index);
-		buffer.Puts(ExpandEscape(iter->name));
-		buffer.Putc(' ');
+			size_t index1 = index;
+			for (size_t i = 0; i < numDigit && index1; ++i) {
+				s[numDigit-i] = (index1 % 10) + '0';
+				index1 /= 10;
+			}
+			html.DrawString(pad, s);
+
+			html.DrawString(pad, iter->name);
+
+			html.DrawChar(pad, ' ');
 		
-		for (size_t i = 0; i < max_length-GetStringWidth(iter->name); i++) {
-			buffer.Putc(' ');
+			for (size_t i = 0; i < max_length-GetStringWidth(iter->name); i++) {
+				html.DrawChar(pad, ' ');
+			}
+
+		    			// End <A CALL=...> tag
+			html.DrawAEnd(pad);
+
+			index++;
 		}
-
-		    			// End <A NAME=...> tag
-    		buffer.Puts("</A>");
-		index++;
-		curCol = 3+numDigit+max_length;
 	}
-
-	buffer.Puts("</PRE>");		// End preformatted text
-
-	if (curCol != 0)		// Not end with new line
-		NewLineUpdateRowCol();	// One more line needed
 }
 
 /*************************************************************************
@@ -771,6 +837,8 @@ void	CD(const string &linktext)
 
 	Buffer	buffer;
 
+	HyperDraw *drawObj = 0;
+
 	if (saveCwd.size())				// Current directory is available
 		FindDir(saveCwd, false);
 
@@ -779,8 +847,8 @@ void	CD(const string &linktext)
 	
 	GenerateHTMLStrings();
 
+	list<MatchInfo>	match_list;
 	if (linktext.size()) {				// kcd DIR
-		list<MatchInfo>	match_list;
 							// We add current
 							// directory into the mix
 							// to aid cycling through
@@ -804,7 +872,7 @@ void	CD(const string &linktext)
 							// Current directory is not
 							// need here.  Remove it.
 			match_list.remove_if(MatchInfoIsCurDir());
-			GenerateDirListHTML(match_list, buffer);
+			drawObj = new HyperListDraw(match_list);
 			
 			hyperInitPos = "";
 			hyperFindLink = "";
@@ -816,7 +884,7 @@ void	CD(const string &linktext)
 		}
 	}
 	else {						// Browse dir tree
-		GenerateHTML(buffer);
+		drawObj = new HyperTreeDraw();
 
 		hyperInitPos = saveCwd;
 		hyperFindLink = "";
@@ -830,7 +898,7 @@ void	CD(const string &linktext)
 	}
 #endif
 
-	HyperDocument		html(buffer.GetPtr(), buffer.GetSize());
+	HyperDocument		*html = new HyperDocument(drawObj);
 
 	{
 						// Display
@@ -842,6 +910,8 @@ void	CD(const string &linktext)
 				throw ErrorScreenTooSmall();
 			}
 			else {
+				init_entity_table();
+
 				khGeometryManager	geoMan(modeHVScroll);
 				khScreenManager		scrnMan(geoMan);
 
@@ -858,7 +928,7 @@ void	CD(const string &linktext)
 							// Use pre-computed 
 							// document size
 				khHyperWindowWithFind	hyper(scrnMan, idHyperWindow,
-							      html, findBar, urlBar,
+							      *html, findBar, urlBar,
 							      hscroll, vscroll, hyperInitPos,
 							      1, htmlRow, htmlCol);
 
