@@ -1,6 +1,6 @@
 /*
 	khdoc.cc	Hypertext Document Classes
-	Copyright (c) 1996-9,2000,2001 Kriang Lerdsuwanakij
+	Copyright (c) 1996-9,2000-2 Kriang Lerdsuwanakij
 	email:		lerdsuwa@users.sourceforge.net
 
 	This program is free software; you can redistribute it and/or modify
@@ -83,10 +83,12 @@
 #endif
 #include <sys/stat.h>
 
-#ifdef HAVE_WCHAR_H
-# define __USE_XOPEN
-# include CXX__HEADER_cwchar
-# include CXX__HEADER_cwctype
+#ifdef USE_UTF8_MODE
+# ifdef HAVE_WCHAR_H
+#  define __USE_XOPEN
+#  include CXX__HEADER_cwchar
+#  include CXX__HEADER_cwctype
+# endif
 #endif
 
 // #define DEBUG_FORMAT_TEXT
@@ -129,16 +131,16 @@ inline void	ThrowErrorMessage(const string &/*msg*/, char * /*buffer*/, int /*fr
 
 // Find line number given index of buffer
 
-int	GetSourceLine(char *buffer, int index)
+size_t	GetSourceLine(char *buffer, size_t index)
 {
-	int	lineNo = 1;
-	for (int i = 0; i <= index; i++)
+	size_t	lineNo = 1;
+	for (size_t i = 0; i <= index; i++)
 		if (buffer[i] == '\n')		// End of line found
 			lineNo++;
 	return lineNo;
 }
 
-void	ThrowErrorMessage(const string &msg, char *buffer, int from, int length)
+void	ThrowErrorMessage(const string &msg, char *buffer, size_t from, size_t length)
 {
 	ostrstream bufstr;
 	gtout(bufstr, msg) << GetSourceLine(buffer, from);
@@ -146,7 +148,7 @@ void	ThrowErrorMessage(const string &msg, char *buffer, int from, int length)
 	throw ErrorGenericSyntax(bufstr);
 }
 
-void	ThrowErrorTag1(const string &tag, const string &msg, char *buffer, int from, int length)
+void	ThrowErrorTag1(const string &tag, const string &msg, char *buffer, size_t from, size_t length)
 {
 	ostrstream bufstr;
 	gtout(bufstr, msg) << tag << GetSourceLine(buffer, from);
@@ -154,7 +156,7 @@ void	ThrowErrorTag1(const string &tag, const string &msg, char *buffer, int from
 	throw ErrorGenericSyntax(bufstr);
 }
 
-void	ThrowErrorTag2(const string &tag1, const string &tag2, const string &msg, char *buffer, int from, int length)
+void	ThrowErrorTag2(const string &tag1, const string &tag2, const string &msg, char *buffer, size_t from, size_t length)
 {
 	ostrstream bufstr;
 	gtout(bufstr, msg) << tag1 << tag2 << GetSourceLine(buffer, from);
@@ -162,32 +164,32 @@ void	ThrowErrorTag2(const string &tag1, const string &tag2, const string &msg, c
 	throw ErrorGenericSyntax(bufstr);
 }
 
-void	ThrowErrorTagInside(const string &tag1, const string &tag2, char *buffer, int from, int length)
+void	ThrowErrorTagInside(const string &tag1, const string &tag2, char *buffer, size_t from, size_t length)
 {
 	ThrowErrorTag2(tag1, tag2, _("%$ tag inside %$ tag at line %$\n"), buffer, from, length);
 }
 
-void	ThrowErrorNestedTag(const string &tag, char *buffer, int from, int length)
+void	ThrowErrorNestedTag(const string &tag, char *buffer, size_t from, size_t length)
 {
 	ThrowErrorTag1(tag, _("nested %$ tag at line %$\n"), buffer, from, length);
 }
 
-void	ThrowErrorTagWithout(const string &tag1, const string &tag2, char *buffer, int from, int length)
+void	ThrowErrorTagWithout(const string &tag1, const string &tag2, char *buffer, size_t from, size_t length)
 {
 	ThrowErrorTag2(tag1, tag2, _("%$ without %$ at line %$\n"), buffer, from, length);
 }
 
-void	ThrowErrorSymbolNeeded(const string &tag1, const string &tag2, char *buffer, int from, int length)
+void	ThrowErrorSymbolNeeded(const string &tag1, const string &tag2, char *buffer, size_t from, size_t length)
 {
 	ThrowErrorTag2(tag1, tag2, _("%$ needed %$ at line %$\n"), buffer, from, length);
 }
 
-void	ThrowErrorEndOfFile(const string &tag, char *buffer, int from, int length)
+void	ThrowErrorEndOfFile(const string &tag, char *buffer, size_t from, size_t length)
 {
 	ThrowErrorTag1(tag, _("unexpected end of file inside %$ at line %$\n"), buffer, from, length);
 }
 
-void	ThrowErrorCannotFind(const string &tag, char *buffer, int from, int length)
+void	ThrowErrorCannotFind(const string &tag, char *buffer, size_t from, size_t length)
 {
 	ThrowErrorTag1(tag, _("cannot find %$ at line %$\n"), buffer, from, length);
 }
@@ -195,172 +197,11 @@ void	ThrowErrorCannotFind(const string &tag, char *buffer, int from, int length)
 #endif	/* TRIM_NO_CHECK */
 
 /*************************************************************************
-	Entity info
+	HyperDraw class
 *************************************************************************/
 
-struct	entity_info {
-	const char	*name;
-	size_t		length;
-	chtype		code;
-};
-
-entity_info entity_table[] = {
-	{ "lt", 2, '<' },
-	{ "gt", 2, '>' },
-	{ "amp", 3, '&' },
-	{ "quot", 4, '\"' },
-
-#ifndef TRIM_NO_LINEDRAW
-	{ "boxur", 5, 0 },			// 4
-	{ "boxdr", 5, 0 },
-	{ "boxdl", 5, 0 },
-	{ "boxul", 5, 0 },
-
-	{ "boxvl", 5, 0 },
-	{ "boxvr", 5, 0 },
-	{ "boxhu", 5, 0 },
-	{ "boxhd", 5, 0 },
-
-	{ "boxh", 4, 0 },
-	{ "boxv", 4, 0 },
-
-	{ "rarr", 4, 0 },
-	{ "larr", 4, 0 },
-	{ "darr", 4, 0 },
-	{ "uarr", 4, 0 },
-#endif
-
-#ifndef TRIM_NO_LATIN
-	{ "iexcl", 5, 161 },
-	{ "cent", 4, 162 },
-	{ "pound", 5, 163 },
-	{ "curren", 6, 164 },
-	{ "yen", 3, 165 },
-	{ "brvbar", 6, 166 },
-	{ "sect", 4, 167 },
-	{ "uml", 3, 168 },
-	{ "copy", 4, 169 },
-	{ "ordf", 4, 170 },
-	{ "laquo", 5, 171 },
-	{ "not", 3, 172 },
-	{ "shy", 3, 173 },
-	{ "reg", 3, 174 },
-	{ "macr", 3, 175 },
-	{ "deg", 3, 176 },
-	{ "plusmn", 6, 177 },
-	{ "sup2", 4, 178 },
-	{ "sup3", 4, 179 },
-	{ "acute", 5, 180 },
-	{ "micro", 5, 181 },
-	{ "para", 4, 182 },
-	{ "middot", 6, 183 },
-	{ "cedil", 5, 184 },
-	{ "sup1", 4, 185 },
-	{ "ordm", 4, 186 },
-	{ "raquo", 5, 187 },
-	{ "frac14", 6, 188 },
-	{ "frac12", 6, 189 },
-	{ "frac34", 6, 190 },
-	{ "iquest", 6, 191 },
-
-	{ "Agrave", 6, 192 },
-	{ "Acute", 5, 193 },
-	{ "Acirc", 5, 194 },
-	{ "Atilde", 6, 195 },
-	{ "Auml", 4, 196 },
-	{ "Aring", 5, 197 },
-	{ "AElig", 5, 198 },
-	{ "Ccedil", 6, 199 },
-	{ "Egrave", 6, 200 },
-	{ "Eacute", 6, 201 },
-	{ "Ecirc", 5, 202 },
-	{ "Euml", 4, 203 },
-	{ "Igrave", 6, 204 },
-	{ "Iacute", 6, 205 },
-	{ "Icirc", 5, 206 },
-	{ "Iuml", 4, 207 },
-	{ "ETH", 3, 208 },
-	{ "Ntilde", 6, 209 },
-	{ "Ograve", 6, 210 },
-	{ "Oacute", 6, 211 },
-	{ "Ocirc", 5, 212 },
-	{ "Otilde", 6, 213 },
-	{ "Ouml", 4, 214 },
-	{ "times", 5, 215 },
-	{ "Oslash", 6, 216 },
-	{ "Ugrave", 6, 217 },
-	{ "Uacute", 6, 218 },
-	{ "Ucirc", 5, 219 },
-	{ "Uuml", 4, 220 },
-	{ "Yacute", 6, 221 },
-	{ "THORN", 5, 222 },
-
-	{ "szlig", 5, 223 },
-	{ "agrave", 6, 224 },
-	{ "aacute", 6, 225 },
-	{ "acirc", 5, 226 },
-	{ "atilde", 6, 227 },
-	{ "auml", 4, 228 },
-	{ "aring", 5, 229 },
-
-	{ "aelig", 5, 230 },
-	{ "ccedil", 6, 231 },
-	{ "egrave", 6, 232 },
-	{ "eacute", 6, 233 },
-	{ "ecirc", 5, 234 },
-	{ "euml", 4, 235 },
-	{ "igrave", 6, 236 },
-	{ "iacute", 6, 237 },
-	{ "icirc", 5, 238 },
-	{ "iuml", 4, 239 },
-	{ "eth", 3, 240 },
-	{ "ntilde", 6, 241 },
-	{ "ograve", 6, 242 },
-	{ "oacute", 6, 243 },
-	{ "ocirc", 5, 244 },
-	{ "otilde", 6, 245 },
-	{ "ouml", 4, 246 },
-	{ "divide", 6, 247 },
-	{ "oslash", 6, 248 },
-	{ "ugrave", 6, 249 },
-	{ "uacute", 6, 250 },
-	{ "ucirc", 5, 251 },
-	{ "uuml", 4, 252 },
-	{ "yacute", 6, 253 },
-	{ "thorn", 5, 254 },
-	{ "yuml", 4, 255 },
-#endif
-
-	{ 0, 0, 0 }
-};
-
-void	init_entity_table()
+HyperDraw::~HyperDraw()
 {
-	static bool init = false;
-	if (!init) {
-#ifndef TRIM_NO_LINEDRAW
-				// Need to initialized here since ACS_*
-				// is not a compile time constant.
-		entity_table[4].code = ACSFallBack(ACS_LLCORNER);
-		entity_table[5].code = ACSFallBack(ACS_ULCORNER);
-		entity_table[6].code = ACSFallBack(ACS_URCORNER);
-		entity_table[7].code = ACSFallBack(ACS_LRCORNER);
-
-		entity_table[8].code = ACSFallBack(ACS_RTEE);
-		entity_table[9].code = ACSFallBack(ACS_LTEE);
-		entity_table[10].code = ACSFallBack(ACS_BTEE);
-		entity_table[11].code = ACSFallBack(ACS_TTEE);
-
-		entity_table[12].code = ACSFallBack(ACS_HLINE);
-		entity_table[13].code = ACSFallBack(ACS_VLINE);
-
-		entity_table[14].code = ACSFallBack(ACS_RARROW);
-		entity_table[15].code = ACSFallBack(ACS_LARROW);
-		entity_table[16].code = ACSFallBack(ACS_DARROW);
-		entity_table[17].code = ACSFallBack(ACS_UARROW);
-#endif
-		init = true;
-	}
 }
 
 /*************************************************************************
@@ -368,8 +209,8 @@ void	init_entity_table()
 *************************************************************************/
 
 // Static member
-const int	HyperDocument::ALLOCSIZE;
-const int	HyperDocument::ALLOCINC;
+const size_t	HyperDocument::ALLOCSIZE;
+const size_t	HyperDocument::ALLOCINC;
 
 void	HyperDocument::Init()		// Used by constructors
 {
@@ -412,29 +253,45 @@ void	HyperDocument::ReInitFormat()
 // Load document from buffer
 //------------------------------------------------------------------------
 
-HyperDocument::HyperDocument(char *buffer_, int bufferSize_)
+HyperDocument::HyperDocument(char *buffer_, size_t bufferSize_)
 {
 	Init();
 	InternalLoadDocument(buffer_, bufferSize_);
 }
 
-void	HyperDocument::LoadDocument(char *buffer_, int bufferSize_)
+void	HyperDocument::LoadDocument(char *buffer_, size_t bufferSize_)
 {
 	ReInit();
 	InternalLoadDocument(buffer_, bufferSize_);
 }
 
-void	HyperDocument::InternalLoadDocument(char *buffer_, int bufferSize_)
+void	HyperDocument::InternalLoadDocument(char *buffer_, size_t bufferSize_)
 {
 	buffer = buffer_;
 	bufferSize = bufferSize_;
 	isMyBuffer = false;
+	drawObj = 0;
+}
+
+//------------------------------------------------------------------------
+// Empty document for run-time generation
+//------------------------------------------------------------------------
+
+HyperDocument::HyperDocument(HyperDraw *drawObj_)
+{
+	Init();
+
+	buffer = 0;
+	bufferSize = 0;
+	isMyBuffer = true;
+	drawObj = drawObj_;
 }
 
 //------------------------------------------------------------------------
 // Load document from the file given
 //------------------------------------------------------------------------
 
+#ifndef TRIM_NO_DOC_FILE
 HyperDocument::HyperDocument(const string &fileName)
 {
 	Init();
@@ -451,9 +308,10 @@ void	HyperDocument::InternalLoadDocument(const string &fileName)
 {
 	isMyBuffer = true;
 	buffer = NULL;				// Not allocated yet
+	drawObj = 0;
 
 	gzFile	fileGZ1 = NULL, fileGZ2 = NULL;	// Avoid warning
-	int	isDir1 = 1, isDir2 = 1;		// Default to 1 when
+	bool	isDir1 = true, isDir2 = true;	// Default to true when
 						// file not found
 	struct stat	fileStat;
 
@@ -513,11 +371,13 @@ void	HyperDocument::InternalLoadDocument(const string &fileName)
 	if (fileGZ2)
 		gzclose(fileGZ2);
 }
+#endif	/* TRIM_NO_DOC_FILE */
 
 //------------------------------------------------------------------------
 // Load document from opened file stream
 //------------------------------------------------------------------------
 
+#ifndef TRIM_NO_DOC_STREAM
 HyperDocument::HyperDocument(FILE *file)
 {
 	Init();
@@ -534,6 +394,7 @@ void	HyperDocument::InternalLoadDocument(FILE *file)
 {
 	isMyBuffer = true;
 	buffer = NULL;			// Not allocated yet
+	drawObj = 0;
 
 	gzFile	fileGZ;
 
@@ -548,32 +409,39 @@ void	HyperDocument::InternalLoadDocument(FILE *file)
 	}
 	gzclose(fileGZ);
 }
+#endif	/* TRIM_NO_DOC_STREAM */
 
 //------------------------------------------------------------------------
 // File I/O
 //------------------------------------------------------------------------
 
+#if ! (defined(TRIM_NO_DOC_STREAM) && defined(TRIM_NO_DOC_FILE))
 void	HyperDocument::ReadFile(gzFile file)
 {
-	int	bufferPtr = 0;
-	int	byteRead;
-	int	readSize = ALLOCSIZE;
+	size_t	bufferPtr = 0;
+	int	byteRead;		// Must be int, gzread returns -1
+					// upon failure
+	size_t	readSize = ALLOCSIZE;
 
 	if (buffer) {
 		delete buffer;
 	}
 
-	buffer = static_cast<char *>(malloc(ALLOCSIZE));	// Allocate buffer
+	buffer = static_cast<char *>(malloc(ALLOCSIZE));// Allocate buffer
+							// Need malloc
+							// because of realloc
+							// below
 	bufferSize = ALLOCSIZE;
 	if (buffer == NULL)
 		throw BAD_ALLOC();
-		
+
 	while((byteRead = gzread(file, buffer+bufferPtr, readSize)) ==
-	      readSize) {
+	      static_cast<int>(readSize)) {
 		char *newBuffer;
 		bufferPtr += byteRead;		// Next read position
 		bufferSize += ALLOCINC;		// New buffer size
 		readSize = ALLOCINC;		// Number of bytes read
+						// Expand memory block
 		newBuffer = static_cast<char *>(realloc(buffer, bufferSize));
 		if (newBuffer == NULL)
 			throw BAD_ALLOC();
@@ -585,6 +453,7 @@ void	HyperDocument::ReadFile(gzFile file)
 
 	bufferSize = bufferSize-readSize+byteRead;
 }
+#endif
 
 //------------------------------------------------------------------------
 // Destructor
@@ -658,11 +527,11 @@ int	HyperDocument::IndentColumn()
 }
 
 #ifdef TRIM_NO_INDENT
-void	HyperDocument::ProcessLI(char * /*buffer*/, int /*length*/, WINDOW * /*pad*/, int /*from*/)
+void	HyperDocument::ProcessLI(char * /*buffer*/, size_t /*length*/, WINDOW * /*pad*/, size_t /*from*/)
 {
 }
 #else	/* TRIM_NO_INDENT */
-void	HyperDocument::ProcessLI(char *buffer, int length, WINDOW *pad, int from)
+void	HyperDocument::ProcessLI(char *buffer, size_t length, WINDOW *pad, size_t from)
 {
 	if (numIndent > 0 && numIndent <= NUM_INDENT) {
 		if (!pad) {
@@ -686,7 +555,7 @@ void	HyperDocument::ProcessLI(char *buffer, int length, WINDOW *pad, int from)
 				case IND_DL:
 					curColumn = indentPos[numIndent-1]-2;
 					wmove(pad, curRow, curColumn);
-					waddch(pad, ACSFallBack(ACS_BULLET));
+					draw_entity(pad, EID_BULLET);
 					break;
 				case IND_OL:
 					{		// Ordered list
@@ -740,11 +609,11 @@ void	HyperDocument::ProcessLI(char *buffer, int length, WINDOW *pad, int from)
 /* pad == NULL indicates that it is called from FindLink(...),
    curRow & curColumn are ignored in this case */
 
-void	HyperDocument::ProcessA(char *buffer, int length, WINDOW *pad, int from,
+void	HyperDocument::ProcessA(char *buffer, size_t length, WINDOW *pad, size_t from,
 		int curRow, int curColumn)
 {
-	int	i = SeekNonSpace(buffer, from, length);
-	int	j = SeekTokenEnd(buffer, i, length);
+	size_t	i = SeekNonSpace(buffer, from, length);
+	size_t	j = SeekTokenEnd(buffer, i, length);
 
 	if (isInA == 1 && procA) {	// Must have </A> for <A HREF=...>
 					// before another <A ...> tag
@@ -780,8 +649,7 @@ void	HyperDocument::ProcessA(char *buffer, int length, WINDOW *pad, int from,
 	}
 
 	j = SeekNonSpace(buffer, j, length);	// Skip spaces
-
-	if (j < 0) {
+	if (j == length) {
 		ThrowErrorEndOfFile(name,
 				    buffer, from-2, length);
 	}
@@ -792,7 +660,7 @@ void	HyperDocument::ProcessA(char *buffer, int length, WINDOW *pad, int from,
 	j++;					// Skip "="
 
 	j = SeekNonSpace(buffer, j, length);	// Skip spaces
-	if (j < 0) {
+	if (j == length) {
 		ThrowErrorEndOfFile(name,
 				    buffer, from-2, length);
 	}
@@ -815,16 +683,18 @@ void	HyperDocument::ProcessA(char *buffer, int length, WINDOW *pad, int from,
 
 	i = j+1;
 
-	if (type == TYPE_EXEC)
+	try {
+		if (type == TYPE_EXEC)
 					// Search closing `
 					// Shell command, can be quoted
-		j = SeekChar('`', buffer, i, length, 2);	// nest support
-	else
-		j = SeekChar('\"', buffer, i, length, 2);	// Search closing "
+			j = SeekChar('`', buffer, i, length, 2);	// nest support
+		else
+			j = SeekChar('\"', buffer, i, length, 2);	// Search closing "
 								// Quote support for kcd
 								// since dir name may 
 								// contain special chars
-	if (j < 0) {
+	}
+	catch (...) {
 		ThrowErrorEndOfFile(name,
 				    buffer, from-2, length);
 	}
@@ -869,16 +739,15 @@ void	HyperDocument::NextLine(WINDOW *pad)
 	IndentLine(pad);			// Indent if necessary
 }
 
-void	HyperDocument::ProcessHypertext(char *buffer, int length, WINDOW *pad)
+void	HyperDocument::ProcessHypertext(char *buffer, size_t length, WINDOW *pad)
 {
 	init_entity_table();
 
-	int	i, j, k;
-	int	ret;
+	size_t	i = 0, j = 0;
+	size_t	k = 0;
 	int	lastCharColumn = 0, lastCharRow = 0;	/* Last char pos */
 	int	lastPreColumn = 0, lastPreRow = 0;	/* Last preformatted pos */
 	int	lastChar = 0;		/* Value ignored */
-	i = 0;
 
 	numIndent = 0;			// Clear indent if this function is
 					// run the second time
@@ -1249,7 +1118,7 @@ void	HyperDocument::ProcessHypertext(char *buffer, int length, WINDOW *pad)
 					for (int curColumn = numColumn/4; curColumn < numColumn*3/4;
 					     curColumn++) {
 						wmove(pad, curRow, curColumn);
-						waddch(pad, ACSFallBack(ACS_HLINE));
+						draw_entity(pad, EID_HLINE);
 					}
 				}
 				NextLine(pad);
@@ -1260,33 +1129,38 @@ void	HyperDocument::ProcessHypertext(char *buffer, int length, WINDOW *pad)
 #endif	/* TRIM_NO_RULE */
 
 			else if (j-i == 5 && CompareStringCase(buffer+i, "TITLE", j-i) == 0) {
-				if ((j = SeekChar('>', buffer, i+5, length)) >= 0) {
-					j++;	/* Skip '>' */
-					k = j;
-					j = SeekCaseString("</TITLE>", buffer, j, length);
-					if (j < 0) {
-						ThrowErrorCannotFind("</TITLE>",
-								     buffer, i-1, length);
-					}
-					if (titleText.size()) {
-						ThrowErrorMessage(_("cannot "
-								    "specify title more "
-								    "than once"),
-								  buffer, i-1, length);
-					}
-						/* Empty <TITLE></TITLE> allowed */
-					if (j-k)
-						titleText = string(buffer+k, j-k);
+				try {
+					j = SeekChar('>', buffer, i+5, length);
 				}
-				else {
+				catch (...) {
 					ThrowErrorCannotFind(_("`>\'"),
 							     buffer, i-1, length);
 				}
+				j++;	/* Skip '>' */
+				k = j;
+				try {
+					j = SeekCaseString("</TITLE>", buffer, j, length);
+				}
+				catch (...) {
+						ThrowErrorCannotFind("</TITLE>",
+								     buffer, i-1, length);
+				}
+				if (titleText.size()) {
+					ThrowErrorMessage(_("cannot "
+							    "specify title more "
+							    "than once"),
+							  buffer, i-1, length);
+				}
+						/* Empty <TITLE></TITLE> allowed */
+				if (j-k)
+					titleText = string(buffer+k, j-k);
 			}
 			else if (CompareStringCase(buffer+i, "!--", 3) == 0 && pad) {
 						/* Seek for "-->" */
-				j = SeekString("-->", buffer, i, length);
-				if (j < 0) {
+				try {
+					j = SeekString("-->", buffer, i, length);
+				}
+				catch (...) {
 					ThrowErrorCannotFind(_("`-->\'"),
 							     buffer, i-1, length);
 				}
@@ -1295,12 +1169,15 @@ void	HyperDocument::ProcessHypertext(char *buffer, int length, WINDOW *pad)
 
 					/* Seek for '>' */
 					/* Can be quoted */
-			ret = SeekChar('>', buffer, j, length, 2);	/* nest support */
-			if (ret < 0) {
+			try {
+				i = SeekChar('>', buffer, j, length, 2);
+							/* nest support */
+			}
+			catch (...) {
 				ThrowErrorCannotFind(_("`>\'"),
 						     buffer, i-1, length);
 			}
-			i = ret+1;
+			i++;
 		}
 		else {
 			chtype	bufChar;	// Enough to cover 0..255
@@ -1318,8 +1195,10 @@ void	HyperDocument::ProcessHypertext(char *buffer, int length, WINDOW *pad)
 				case '&':
 					i++;
 						/* Seek for ';' */
-					j = SeekChar(';', buffer, i, length);
-					if (j < 0) {
+					try {
+						j = SeekChar(';', buffer, i, length);
+					}
+					catch (...) {
 						ThrowErrorCannotFind(_("`;\'"),
 								     buffer, i-1, length);
 					}
@@ -1344,12 +1223,14 @@ void	HyperDocument::ProcessHypertext(char *buffer, int length, WINDOW *pad)
 						for (size_t k = 0; entity_table[k].name; ++k) {
 							if (static_cast<size_t>(j-i) == entity_table[k].length
 							    && strncmp(buffer+i, entity_table[k].name, j-i) == 0) {
+								// FIXME: Handle wide/combining entity
 								bufChar = entity_table[k].code;
 								break;
 							}
 						}
 						//if (bufChar == 0)	// Unknown escape, print as a bullet
-						//	bufChar = ACSFallBack(ACS_BULLET);
+						//	bufChar = ...get_entity(pad, EID_BULLET);
+
 					}
 					i = j;	/* i++ appears later... */
 					break;
@@ -1368,8 +1249,15 @@ void	HyperDocument::ProcessHypertext(char *buffer, int length, WINDOW *pad)
 							throw ErrorBadSequence();
 
 						bufChar = static_cast<chtype>(wc);
-						charWidth = wcwidth(wc);
-						i += ret-1;
+
+						// FIXME: Check if printable logic correct?
+						if (iswprint(wc) || wc == ' ') {
+							charWidth = wcwidth(wc);
+							i += ret-1;
+						}
+						else {
+							bufChar = '?';
+						}
 					}
 					else
 #endif
@@ -1720,14 +1608,19 @@ void	HyperDocument::FormatDocument(WINDOW *pad, int row, int col)
 	scrRow = row;
 	scrColumn = col;
 	
-	if (pad) {	// We only need following when we output to pad
-		;
+	if (! drawObj) {
+		if (pad) {	// We only need following when we output to pad
+			;
+		}
+		else {		// We are going to find document dimension
+			numRow = 1;	// Minimum limit - must contain an empty line
+			numColumn = 0;
+		}
+		ProcessHypertext(buffer, bufferSize, pad);
 	}
-	else {		// We are going to find document dimension
-		numRow = 1;	// Minimum limit - must contain an empty line
-		numColumn = 0;
+	else {
+		drawObj->Draw(*this, pad);
 	}
-	ProcessHypertext(buffer, bufferSize, pad);
 }
 
 void	HyperDocument::SetDocumentSize(int docRow, int docCol, int row, int col)
@@ -1748,14 +1641,11 @@ int	HyperDocument::ReadATag()
 	int	currentID = 0;	// ID used to replace rowFrom member of Anchor
 				// It is used to sync between anchorList and
 				// sectionList
-	int	i, j;
-	int	ret;
+	size_t	i = 0, j = 0;
 
 	ReInitFormat();		// Initialize document dependent variables
 				// without freeing buffer
 
-	i = 0;
-
 	while (i < bufferSize) {
 	
 		if (buffer[i] == '<') {		/* New tag found */
@@ -1790,12 +1680,15 @@ int	HyperDocument::ReadATag()
 
 					/* Seek for '>' */
 					/* Can be quoted */
-			ret = SeekChar('>', buffer, j, bufferSize, 2);	/* nest support */
-			if (ret < 0) {
+			try {
+				i = SeekChar('>', buffer, j, bufferSize, 2);
+						/* nest support */
+			}
+			catch (...) {
 				ThrowErrorCannotFind(_("`>\'"),
 						     buffer, i-1, bufferSize);
 			}
-			i = ret+1;
+			i++;
 		}
 		else
 			i++;			// Check next char
@@ -1914,3 +1807,324 @@ Anchor	*HyperDocument::FindNextLink(const string &startSection, const string &fi
 
 	return anchorSelected;
 }
+
+//------------------------------------------------------------------------
+// Draw functions
+//------------------------------------------------------------------------
+
+void	HyperDocument::DrawSetSize(int row, int col)
+{
+	numRow = row;
+	numColumn = col;
+}
+
+void	HyperDocument::DrawBegin(WINDOW *pad)
+{
+	init_entity_table();
+
+	curRow = 0;
+	curColumn = 0;
+	numChar = new short[numRow];
+	for (size_t i = 0; i < numRow; ++i)
+		numChar[i] = numColumn;
+
+	numNestA = 0;
+	numNestB = 0;
+	numNestI = 0;
+
+	wmove(pad, 0, 0);
+}
+
+void	HyperDocument::DrawSetTitle(const char *s)
+{
+	titleText = s;
+}
+
+void	HyperDocument::DrawSetTitle(const string &s)
+{
+	titleText = s;
+}
+
+void	HyperDocument::DrawSetItalic(WINDOW *pad)
+{
+	if (pad) {
+		numNestI++;
+		wattrset(pad, GetNestedAttr());
+	}
+}
+
+void	HyperDocument::DrawClearItalic(WINDOW *pad)
+{
+	if (pad) {
+		if (numNestI > 0)
+			numNestI--;
+		wattrset(pad, GetNestedAttr());
+	}
+}
+
+void	HyperDocument::DrawSetBold(WINDOW *pad)
+{
+	if (pad) {
+		numNestB++;
+		wattrset(pad, GetNestedAttr());
+	}
+}
+
+void	HyperDocument::DrawClearBold(WINDOW *pad)
+{
+	if (pad) {
+		if (numNestB > 0)
+			numNestB--;
+		wattrset(pad, GetNestedAttr());
+	}
+}
+
+void	HyperDocument::DrawCharSub(WINDOW *pad, chtype c)
+{
+	if (c == '\n') {
+		curRow++;
+		curColumn = 0;
+		wmove(pad, curRow, curColumn);
+	}
+	else if (c == '\t' || c == '\r' || c == '\b' || c == '\0')
+		;
+	else {
+		if (isprint(c))
+			waddch(pad, static_cast<unsigned char>(c));
+		else
+			waddch(pad, '?');
+		curColumn++;
+	}
+}
+
+void	HyperDocument::DrawChar(WINDOW *pad, chtype c)
+{
+	if (pad) {
+		if (procA)
+			procA->linkText += static_cast<char>(c & 0xFF);
+
+		DrawCharSub(pad, c);
+	}
+}
+
+#ifdef USE_UTF8_MODE
+void	HyperDocument::DrawWCharSub(WINDOW *pad, wchar_t wc)
+{
+	if (wc == '\n') {
+		curRow++;
+		curColumn = 0;
+		wmove(pad, curRow, curColumn);
+	}
+	else if (wc == '\t' || wc == '\r' || wc == '\b' || wc == '\0')
+		;
+	else {
+		if (iswprint(wc)) {
+			int w = wcwidth(wc);
+			if (w == 0 && curColumn) {
+				cchar_t	cc;
+				wmove(pad, curRow, curColumn-1);
+				win_wch(pad, &cc);
+
+				bool done = false;
+				for (size_t i = 0; i < CCHARW_MAX; ++i) {
+					if (!cc.chars[i]) {
+						cc.chars[i] = wc;
+						done = true;
+						break;
+					}
+				}
+
+				if (done) {
+					wmove(pad, curRow, curColumn-1);
+					wadd_wch(pad, &cc);
+				}
+				else {
+					wmove(pad, curRow, curColumn);
+					waddnwstr(pad, &wc, 1);
+				}
+			}
+			else {
+				waddnwstr(pad, &wc, 1);
+			}
+			curColumn += wcwidth(wc);
+		}
+		else {
+			waddch(pad, '?');
+			curColumn++;
+		}
+	}
+}
+#endif /* USE_UTF8_MODE */
+
+void	HyperDocument::DrawEntityChar(WINDOW *pad, entity_id id)
+{
+	if (pad) {
+		curColumn++;
+		draw_entity(pad, id);
+	}
+}
+
+void	HyperDocument::DrawString(WINDOW *pad, const char *s)
+{
+	if (pad) {
+		if (procA)
+			procA->linkText += s;
+
+#ifdef USE_UTF8_MODE
+// FIXME: Handle wide/combining chars, UTF-8 encoding
+//		if (0) {
+		if (IsUTF8Mode()) {
+			size_t length = strlen(s);
+
+			for (size_t i = 0; s[i] != '\0'; ) {
+				wchar_t wc;
+				static mbstate_t state;
+				memset(&state, 0, sizeof(mbstate_t));
+				int ret = mbrtowc(&wc, s+i, length-i, &state);
+				if (ret <= 0)
+					throw ErrorBadSequence();
+
+						// FIXME: Check if printable logic correct?
+				if (iswprint(wc) || wc == ' ') {
+					DrawWCharSub(pad, wc);
+				}
+				else {
+					DrawCharSub(pad, '?');
+				}
+				i += ret;
+			}
+		}
+		else
+#endif
+			for (size_t i = 0; s[i] != '\0'; ++i)
+				DrawCharSub(pad, s[i]);
+	}
+}
+
+void	HyperDocument::DrawString(WINDOW *pad, const string &s)
+{
+	DrawString(pad, s.c_str());
+}
+
+void	HyperDocument::DrawAName(WINDOW *pad, const char *s)
+{
+	if (pad) {
+		Anchor *proc = new Anchor(s);
+
+		proc->type = TYPE_SECTION;
+		proc->rowFrom = curRow;
+		proc->colFrom = curColumn;
+
+		proc->rowTo = curRow;		// Don't care where </A> is
+		proc->colTo = curColumn;
+
+		sectionList.push_back(sptr<Anchor>(proc));
+	}
+}
+
+void	HyperDocument::DrawAName(WINDOW *pad, const string &s)
+{
+	if (pad) {
+		Anchor *proc = new Anchor(s);
+
+		proc->type = TYPE_SECTION;
+		proc->rowFrom = curRow;
+		proc->colFrom = curColumn;
+
+		proc->rowTo = curRow;		// Don't care where </A> is
+		proc->colTo = curColumn;
+
+		sectionList.push_back(sptr<Anchor>(proc));
+	}
+}
+
+void	HyperDocument::DrawACallBegin(WINDOW *pad, const char *s)
+{
+	if (pad) {
+		numNestA++;
+		wattrset(pad, GetNestedAttr());
+
+		procA = new Anchor(s);
+
+		procA->type = TYPE_CALL;
+		procA->rowFrom = curRow;
+		procA->colFrom = curColumn;
+
+		procA->rowTo = -1;
+		procA->colTo = -1;
+
+		anchorList.push_back(sptr<Anchor>(procA));
+	}
+}
+
+void	HyperDocument::DrawACallBegin(WINDOW *pad, const string &s)
+{
+	if (pad) {
+		numNestA++;
+		wattrset(pad, GetNestedAttr());
+
+		procA = new Anchor(s);
+
+		procA->type = TYPE_CALL;
+		procA->rowFrom = curRow;
+		procA->colFrom = curColumn;
+
+		procA->rowTo = -1;
+		procA->colTo = -1;
+
+		anchorList.push_back(sptr<Anchor>(procA));
+	}
+}
+
+void	HyperDocument::DrawAHrefBegin(WINDOW *pad, const char *s)
+{
+	if (pad) {
+		numNestA++;
+		wattrset(pad, GetNestedAttr());
+
+		procA = new Anchor(s);
+
+		procA->type = TYPE_GOTO;
+		procA->rowFrom = curRow;
+		procA->colFrom = curColumn;
+
+		procA->rowTo = -1;
+		procA->colTo = -1;
+
+		anchorList.push_back(sptr<Anchor>(procA));
+	}
+}
+
+void	HyperDocument::DrawAHrefBegin(WINDOW *pad, const string &s)
+{
+	if (pad) {
+		numNestA++;
+		wattrset(pad, GetNestedAttr());
+
+		procA = new Anchor(s);
+
+		procA->type = TYPE_GOTO;
+		procA->rowFrom = curRow;
+		procA->colFrom = curColumn;
+
+		procA->rowTo = -1;
+		procA->colTo = -1;
+
+		anchorList.push_back(sptr<Anchor>(procA));
+	}
+}
+
+void	HyperDocument::DrawAEnd(WINDOW *pad)
+{
+	if (pad) {
+		if (numNestA > 0) {
+			procA->rowTo = curRow;
+			procA->colTo = curColumn-1;	/* Prev column */
+
+			numNestA--;
+			wattrset(pad, GetNestedAttr());
+
+			procA = 0;
+		}
+	}
+}
