@@ -1,6 +1,6 @@
 /*
 	khwin.cc	Hypertext Window
-	Copyright (c) 1997-8,2000 Kriang Lerdsuwanakij
+	Copyright (c) 1997-8,2000,2002 Kriang Lerdsuwanakij
 	email:		lerdsuwa@users.sourceforge.net
 
 	This program is free software; you can redistribute it and/or modify
@@ -223,7 +223,7 @@ ATTR_TYPE	normalA, linkA, highlightA, barA, urlA, headerA, boldA,
 		italicA, markA, linkBoldA, linkItalicA, highlightBoldA,
 		highlightItalicA;
 ATTR_TYPE	scrollArrowA, scrollBarA, scrollBlockA;
-chtype		scrollBarC, scrollBlockC;
+entity_id	scrollBarID, scrollBlockID;
 
 void	khScreenManager::RequestResize()
 {
@@ -313,11 +313,11 @@ khScreenManager::khScreenManager(NCGeometryManagerBase &geoMan_) :
 	InitAttr(29, kcdConfig.cfgAttrHighlightBold, highlightBoldA);
 	InitAttr(31, kcdConfig.cfgAttrHighlightItalic, highlightItalicA);
 
-	scrollBlockC = ' ';
-	if (isACSFallBack)
-		scrollBarC = ' ';
+	scrollBlockID = EID_SPACE;
+	if (isACSFallBack)		// No special symbols
+		scrollBarID = EID_SPACE;
 	else
-		scrollBarC = ACSFallBack(ACS_CKBOARD);
+		scrollBarID = EID_CKBOARD;
 }
 
 
@@ -463,7 +463,6 @@ void	HyperWindow::Highlight()
 	int	curR, curC;
 	int	row = curRow+padRow, column = curColumn+padColumn;
 	int	prev = 0, next = 0;	// Linked-list traverse direction
-	chtype	c, chr, attr;
 
 	hscroll.SetPosition(padColumn);
 	vscroll.SetPosition(padRow);
@@ -502,22 +501,47 @@ void	HyperWindow::Highlight()
 			}
 			else {
 								// Replace attribute
-								// FIXME: Unicode
-				c = winch(win);
-				attr = c & ~(A_CHARTEXT|A_ALTCHARSET);
-				chr = c & (A_CHARTEXT|A_ALTCHARSET);
-				if (attr == static_cast<chtype>(highlightA)) {
-					wattrset(win, linkA);
-				}
-				else if (attr == static_cast<chtype>(highlightBoldA)) {
-					wattrset(win, linkBoldA);
-				}
-				else if (attr == static_cast<chtype>(highlightItalicA)) {
-					wattrset(win, linkItalicA);
-				}
+#ifdef USE_UTF8_MODE
+				if (IsUTF8Mode()) {
+					cchar_t cc;
+					win_wch(win, &cc);
+					chtype	attr = cc.attr;
+					if (attr == static_cast<chtype>(highlightA)) {
+						wattrset(win, linkA);
+						cc.attr = linkA;
+					}
+					else if (attr == static_cast<chtype>(highlightBoldA)) {
+						wattrset(win, linkBoldA);
+						cc.attr = linkBoldA;
+					}
+					else if (attr == static_cast<chtype>(highlightItalicA)) {
+						wattrset(win, linkItalicA);
+						cc.attr = linkItalicA;
+					}
 
-				waddch(win, chr);
-				curC++;
+					wadd_wch(win, &cc);
+					curC++;
+				}
+				else
+#endif
+				{
+					chtype	c, chr, attr;
+					c = winch(win);
+					attr = c & ~(A_CHARTEXT|A_ALTCHARSET);
+					chr = c & (A_CHARTEXT|A_ALTCHARSET);
+					if (attr == static_cast<chtype>(highlightA)) {
+						wattrset(win, linkA);
+					}
+					else if (attr == static_cast<chtype>(highlightBoldA)) {
+						wattrset(win, linkBoldA);
+					}
+					else if (attr == static_cast<chtype>(highlightItalicA)) {
+						wattrset(win, linkItalicA);
+					}
+
+					waddch(win, chr);
+					curC++;
+				}
 			}
 		}
 		prevHighlight = NULL;
@@ -553,21 +577,47 @@ void	HyperWindow::Highlight()
 				else {
 									// Replace attribute
 									// Fixme Unicode
-					c = winch(win);
-					attr = c & ~(A_CHARTEXT|A_ALTCHARSET);
-					chr = c & (A_CHARTEXT|A_ALTCHARSET);
-					if (attr == static_cast<chtype>(linkA)) {
-						wattrset(win, highlightA);
-					}
-					else if (attr == static_cast<chtype>(linkBoldA)) {
-						wattrset(win, highlightBoldA);
-					}
-					else if (attr == static_cast<chtype>(linkItalicA)) {
-						wattrset(win, highlightItalicA);
-					}
+#ifdef USE_UTF8_MODE
+					if (IsUTF8Mode()) {
+						cchar_t cc;
+						win_wch(win, &cc);
+						chtype	attr = cc.attr;
+						if (attr == static_cast<chtype>(linkA)) {
+							wattrset(win, highlightA);
+							cc.attr = highlightA;
+						}
+						else if (attr == static_cast<chtype>(linkBoldA)) {
+							wattrset(win, highlightBoldA);
+							cc.attr = highlightBoldA;
+						}
+						else if (attr == static_cast<chtype>(linkItalicA)) {
+							wattrset(win, highlightItalicA);
+							cc.attr = highlightItalicA;
+						}
 
-					waddch(win, chr);
-					curC++;
+						wadd_wch(win, &cc);
+						curC++;
+					}
+					else
+#endif
+					{
+						chtype	c, chr, attr;
+						c = winch(win);
+						attr = c & ~(A_CHARTEXT|A_ALTCHARSET);
+						chr = c & (A_CHARTEXT|A_ALTCHARSET);
+						if (attr == static_cast<chtype>(linkA)) {
+							wattrset(win, highlightA);
+						}
+						else if (attr == static_cast<chtype>(linkBoldA)) {
+							wattrset(win, highlightBoldA);
+						}
+						else if (attr == static_cast<chtype>(linkItalicA)) {
+							wattrset(win, highlightItalicA);
+						}
+
+						waddch(win, chr);
+						curC++;
+					}
 				}
 			}
 			url.SetTitle((*curAnchorIter)->name);
@@ -735,51 +785,11 @@ void	HyperWindow::PrepareDisplay(bool sameDocument)
 	scrnMan.RequestRestCursor();
 }
 
-int	HyperWindow::MatchSubString(iterType &curA, const string &findText)
+bool	HyperWindow::MatchSubString(iterType &curA, const string &findText)
 {
-#ifdef USE_UTF8_MODE
-	static mbstate_t state;
-#endif
-
-	int	x = (*curA)->colFrom, y = (*curA)->rowFrom;
-	chtype	c;
-
-	string linkText;
-	while (y < (*curA)->rowTo ||
-	       (y == (*curA)->rowTo && x <= (*curA)->colTo)) {
-	       	wmove(win, y, x);
-	       				// Extract char code
-					// FIXME: Unicode
-	       	c = (winch(win) & (A_CHARTEXT|A_ALTCHARSET));
-
-		if (c & A_ALTCHARSET)
-					// Special symbols
-					// Should not match
-					// anything
-			linkText += static_cast<char>(1);
-#ifdef USE_UTF8_MODE
-		else if (IsUTF8Mode() && c > 0x7F) {
-			Buffer	buf(1000);
-			memset(&state, 0, sizeof(mbstate_t));
-			if (wcrtomb(buf.GetPtr(), c, &state)
-			    == static_cast<size_t>(-1))
-				throw ErrorBadSequence();
-			linkText += buf.GetPtr();
-		}
-#endif
-		else
-			linkText += static_cast<char>(c);
-		x++;
-		if (x >= GetNumChar(y)) {
-			y++;
-			x = 0;
-		}
-	}
-
-	if (linkText.find(findText) != string::npos) {
-		return 1;	// Found
-	}
-	return 0;		// Not found
+	if ((*curA)->linkText.find(findText) != string::npos)
+		return true;	// Found
+	return false;		// Not found
 }
 
 void	HyperWindow::DoResize()
@@ -897,6 +907,57 @@ void	HyperWindow::DoUpdate()
 		wnoutrefresh(blankWin);
 }
 
+void	HyperWindow::FitHighlightHorizontal(iterType &newA)
+{
+	// Six cases to consider
+	// Legends: |      | = screen boundary, **** = highlight
+	// 1 **|**    |		==>   |****  |
+	// 2 **|***** |		==>  *|******|
+	// 3 **|******|*	==> **|******|*
+	// 4   | **** |		=>    | **** |
+	// 5   |    **|**	==>   |  ****|
+	// 6   | *****|**	==>   |******|*
+
+	if (curColumn < 0) {
+		if ((*newA)->colTo-(*newA)->colFrom <= GetCol()-1) {
+				// Case 1
+				// Scroll right
+			padColumn += curColumn;
+			curColumn = 0;
+		}
+		else if (curColumn+(*newA)->colTo-(*newA)->colFrom <= GetCol()-1) {
+				// Case 2
+			int diff = GetCol()-1
+				   -curColumn-(*newA)->colTo+(*newA)->colFrom;
+			padColumn += diff;
+			curColumn -= diff;
+		}
+		else {
+				// Case 3
+			;
+		}
+	}
+	else {
+		if (curColumn+(*newA)->colTo-(*newA)->colFrom <= GetCol()-1) {
+				// Case 4
+			;
+		}
+		else if ((*newA)->colTo-(*newA)->colFrom <= GetCol()-1) {
+				// Case 5
+				// Scroll left
+			int diff = curColumn+(*newA)->colTo-(*newA)->colFrom
+				   -GetCol()+1;
+			padColumn += diff;
+			curColumn -= diff;
+		}
+		else {
+				// Case 5
+			padColumn += curColumn;
+			curColumn = 0;
+		}
+	}
+}
+
 void	HyperWindow::FitHighlightTop(iterType &newA)
 {
 	if (curRow < 0) {
@@ -904,28 +965,7 @@ void	HyperWindow::FitHighlightTop(iterType &newA)
 		padRow += curRow;
 		curRow = 0;
 	}
-	if (curColumn < 0) {
-					/* Scroll left */
-		padColumn += curColumn;
-		curColumn = 0;
-	}
-	if (curColumn+(*newA)->colTo-(*newA)->colFrom > GetCol()-1) {
-					/* Scroll right */
-		if ((*newA)->colTo-(*newA)->colFrom > GetCol()-1) {
-					// Highlight bar too long - show only
-					// left portion and discard parts 
-					// that go beyond the screen width
-					
-					// Scroll left
-			padColumn += curColumn;
-			curColumn = 0;
-		}
-		else {			// Can fit highlight on the screen
-			padColumn += curColumn+(*newA)->colTo-(*newA)->colFrom
-					-(GetCol()-1);
-			curColumn = GetCol()-1-((*newA)->colTo-(*newA)->colFrom);
-		}
-	}
+	FitHighlightHorizontal(newA);
 }
 
 void	HyperWindow::CenterRowPad(iterType &newA)
@@ -961,28 +1001,7 @@ void	HyperWindow::FitHighlightMiddle(iterType &newA)
 	curRow = (*newA)->rowFrom-padRow;
 	curColumn = (*newA)->colFrom-padColumn;
 
-	if (curColumn < 0) {
-					/* Scroll left */
-		padColumn += curColumn;
-		curColumn = 0;
-	}
-	if (curColumn+(*newA)->colTo-(*newA)->colFrom > GetCol()-1) {
-		if ((*newA)->colTo-(*newA)->colFrom > GetCol()-1) {
-					// Highlight bar too long - show only
-					// left portion and discard parts 
-					// that go beyond the screen width
-					
-					// Scroll left
-			padColumn += curColumn;
-			curColumn = 0;
-		}
-		else {			// Can fit highlight on the screen
-					/* Scroll right */
-			padColumn += curColumn+(*newA)->colTo-(*newA)->colFrom
-					-(GetCol()-1);
-			curColumn = GetCol()-1-((*newA)->colTo-(*newA)->colFrom);
-		}
-	}
+	FitHighlightHorizontal(newA);
 }
 
 void	HyperWindow::FitHighlightBottom(iterType &newA)
@@ -992,29 +1011,8 @@ void	HyperWindow::FitHighlightBottom(iterType &newA)
 		padRow += curRow+(*newA)->rowTo-(*newA)->rowFrom-(GetRow()-1);
 		curRow = GetRow()-1-((*newA)->rowTo-(*newA)->rowFrom);
 	}
-	if (curColumn < 0) {
-					/* Scroll left */
-		padColumn += curColumn;
-		curColumn = 0;
-	}
-	if (curColumn+(*newA)->colTo-(*newA)->colFrom > GetCol()-1) {
-	
-		if ((*newA)->colTo-(*newA)->colFrom > GetCol()-1) {
-					// Highlight bar too long - show only
-					// left portion and discard parts 
-					// that go beyond the screen width
-					
-					// Scroll left
-			padColumn += curColumn;
-			curColumn = 0;
-		}
-		else {			// Can fit highlight on the screen
-					/* Scroll right */
-			padColumn += curColumn+(*newA)->colTo-(*newA)->colFrom
-					-(GetCol()-1);
-			curColumn = GetCol()-1-((*newA)->colTo-(*newA)->colFrom);
-		}
-	}
+
+	FitHighlightHorizontal(newA);
 }
 
 bool	HyperWindow::PosBeforeAnchor(iterType &p, int x, int y)
@@ -1120,7 +1118,7 @@ void	HyperWindow::ProcessKey(int ch)
 				
 			if (anchorSelected->type == TYPE_GOTO) {
 
-				string	startSection, newHyperFile;
+				string	newHyperFile;
 
 				string::size_type s = anchorSelected->name.find('#');
 				if (s != string::npos) {	
@@ -1135,8 +1133,12 @@ void	HyperWindow::ProcessKey(int ch)
 
 						/* Link to another file */
 				if (newHyperFile.size()) {
+#ifndef TRIM_NO_DOC_FILE
 					html.LoadDocument(newHyperFile);
 					PrepareDisplay(false);
+#else
+					;
+#endif
 				}
 				else {	// Link to the same file
 					PrepareDisplay(true);
@@ -1171,7 +1173,7 @@ void	HyperWindow::ProcessKey(int ch)
 			scrnMan.RequestRestCursor();
 		}
 		else		/* Cursor not on any link */
-// FIXME
+				// FIXME: center screen
 			return;
 	}
 	else if (ch == KEY_F(9)) {
@@ -1763,7 +1765,8 @@ void	HyperWindow::KeyboardLoop()
 		blankWin = NULL;
 	}
 
-					// FIXME - delete other windows too
+					// Exit program, don't have
+					// to delete other windows
 
 	clear();			// Clear screen
 	refresh();
@@ -1892,7 +1895,8 @@ void	khHyperWindowWithFind::FindProcessKey(int ch)
 
 void	khHyperWindowWithFind::ProcessKey(int ch)
 {
-	if (isalnum(ch)) {
+				// Non-special keys
+	if (ch <= 0xFF && isalnum(ch)) {
 				// Start find
 
 				// Pass control of keyboard & cursor to
