@@ -188,7 +188,7 @@ void	RecursiveFullScanDir(ScanLog &scan_log, sptr<DirectoryEntry> &d,
 
 						// Do not delete d2 if it is added to
 						// dirTree!
-				sptr<DirectoryEntry> d2(new DirectoryEntry(entry->d_name));
+				sptr<DirectoryEntry> d2(new DirectoryEntry(entry));
 
 				if (d2->IsDir()) {
 
@@ -438,7 +438,7 @@ void	RecursiveSmartScanDir(ScanLog &scan_log, sptr<DirectoryEntry> &d,
 
 						// Do not delete d2 if it is added to
 						// dirTree!
-				sptr<DirectoryEntry> d2(new DirectoryEntry(entry->d_name));
+				sptr<DirectoryEntry> d2(new DirectoryEntry(entry));
 	
 				if (d2->IsDir()) {
 				
